/*
 * Copyright (C) 2023 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *	  Author: Suvendhu Hansa <suvendhu.hansa@iopsys.eu>
 *
 */

#include "times.h"

/* ********** DynamicObj ********** */
DM_MAP_OBJ tDynamicObj[] = {
/* parentobj, nextobject, parameter */
{"Device.", tDeviceObjs, NULL},
{0}
};

/*************************************************************
* ADD & DEL METHODS
**************************************************************/
static int addTimeClient(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char s_name[16] = {0};

	snprintf(s_name, sizeof(s_name), "client_%s", *instance);

	dmuci_add_section("time", "client", &s);
	dmuci_rename_section_by_section(s, s_name);

	dmuci_set_value_by_section(s, "enable", "0");
	dmuci_set_value_by_section(s, "iburst", "1");
	dmuci_set_value_by_section(s, "version", "4");
	dmuci_set_value_by_section(s, "peer", "0");
	dmuci_set_value_by_section(s, "minpoll", "6");
	dmuci_set_value_by_section(s, "maxpoll", "10");
	dmuci_set_value_by_section(s, "mode", "Unicast");

	dmuci_add_section_bbfdm("dmmap_time", "client", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", s_name);
	dmuci_set_value_by_section(dmmap_s, "client_instance", *instance);
	return 0;
}

static int delTimeClient(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
	case DEL_INST:
		dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
		dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
		break;
	case DEL_ALL:
		uci_foreach_sections_safe("time", "client", stmp, s) {
			struct uci_section *dmmap_s = NULL;

			get_dmmap_section_of_config_section("dmmap_time", "client", section_name(s), &dmmap_s);
			dmuci_delete_by_section(dmmap_s, NULL, NULL);

			dmuci_delete_by_section(s, NULL, NULL);
		}
		break;
	}
	return 0;
}

static int addTimeServer(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	bbfdm_set_fault_message(ctx, "The current implementation only supports one 'Server', so adding a new 'Server' is not allowed.");
	return FAULT_9003;
}

static int delTimeServer(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	bbfdm_set_fault_message(ctx, "The current implementation does not allow to delete the 'Server' instance. It can be disabled.");
	return FAULT_9003;
}

/*************************************************************
 * ENTRY METHODS
 *************************************************************/
static int browseTimeClient(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("time", "client", "dmmap_time", &dup_list);
	list_for_each_entry(p, &dup_list, list) {
		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "client_instance", "client_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseTimeServer(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct uci_section *s = is_dmmap_section_exist("dmmap_time", "server");
	if (!s) {
		dmuci_add_section_bbfdm("dmmap_time", "server", &s);
		dmuci_set_value_by_section(s, "server_instance", "1");
	}

	handle_instance(dmctx, parent_node, s, "server_instance", "server_alias");
	DM_LINK_INST_OBJ(dmctx, parent_node, s, "1");
	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
static int get_time_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_option_value_fallback_def("time", "global", "enable", "1");
	return 0;
}

static int set_time_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			ret = FAULT_9007;
		break;
	case VALUESET:
		string_to_bool(value, &b);
		dmuci_set_value("time", "global", "enable", b ? "1" : "0");
		break;
	}

	return ret;
}

static int get_time_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "status", UBUS_ARGS{{"object", "global", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "Status");
	} else {
		*value = "Error";
	}

	return 0;
}

static int get_time_CurrentLocalTime(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return dm_time_format(time(NULL), value);
}

/*#Device.Time.LocalTimeZone!UCI:system/system,@system[0]/timezone*/
static int get_time_LocalTimeZone(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_option_value_string("system", "@system[0]", "timezone", value);
	return 0;
}

static int set_time_LocalTimeZone(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 256, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value("system", "@system[0]", "timezone", value);
			break;
	}
	return 0;
}

static int get_time_ClientNumber(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	unsigned int cnt = get_number_of_entries(ctx, data, instance, browseTimeClient);
	dmasprintf(value, "%u", cnt);
	return 0;
}

static int get_client_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "enable", "1");
	return 0;
}

static int set_client_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			ret = FAULT_9007;
		break;
	case VALUESET:
		string_to_bool(value, &b);
		char *cur = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "enable", "1");
		bool cur_val;
		string_to_bool(cur, &cur_val);
		if (b == cur_val)
			break;

		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "enable", b ? "1" : "0");
		if (b) {
			/* Only one client instance can be enabled at a time */
			struct uci_section *s = NULL, *stmp = NULL;
			uci_foreach_sections_safe("time", "client", stmp, s) {
				char *config_section = section_name(((struct dmmap_dup *)data)->config_section);
				char *uci_section = section_name(s);
				if (DM_STRCMP(uci_section, config_section) != 0)
					dmuci_set_value_by_section(s, "enable", "0");
			}
		}
		break;
	}

	return ret;
}

static int get_client_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	char *sec_name = section_name(((struct dmmap_dup *)data)->config_section);
	dmubus_call("timemngr", "status", UBUS_ARGS{{"object", "client", String}, {"section", sec_name, String}}, 2, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "Status");
	} else {
		*value = "Error";
	}

	return 0;
}

static int get_client_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "client_alias", instance, value);
}

static int set_client_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "client_alias", instance, value);
}

static int get_client_Mode(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "mode", value);
	return 0;
}

static int set_client_Mode(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;
	char *allowed_params[] = {"Unicast", "Broadcast", "Multicast", "Manycast", NULL};

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_string(ctx, value, -1, -1, allowed_params, NULL))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "mode", value);
		break;
	}

	return ret;
}

static int get_client_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "123";
	return 0;
}

static int set_client_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1","65535"}}, 1))
			ret = FAULT_9007;

		if (DM_STRCMP(value, "123") != 0) {
			bbfdm_set_fault_message(ctx, "The current implementation supports only one port which is '123'. So the only allowed value is '123'.");
			ret = FAULT_9007;
		}

		break;
	case VALUESET:
		break;
	}

	return ret;
}

static int get_client_Version(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "version", value);
	return 0;
}

static int set_client_Version(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1", "4"}}, 1))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "version", value);
		break;
	}

	return ret;
}

static int get_client_Servers(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "server", &val);
	*value = dmuci_list_to_string(val, ",");
	return 0;
}

static int set_client_Servers(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	size_t length, i;
	int ret = 0;
	char **arr;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
			ret = FAULT_9007;
		break;
	case VALUESET:
		arr = strsplit(value, ",", &length);
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "server", "");
		for (i = 0; i < length; i++)
			dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, "server", arr[i]);
		break;
	}

	return ret;
}

static int get_client_Peer(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "peer", value);
	return 0;
}

static int set_client_Peer(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			ret = FAULT_9007;
		break;
	case VALUESET:
		string_to_bool(value, &b);
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "peer", b ? "1" : "0");
		break;
	}

	return ret;
}

static int get_client_MinPoll(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "minpoll", "6");
	return 0;
}

static int set_client_MinPoll(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{NULL, NULL}}, 1))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "minpoll", value);
		break;
	}

	return ret;
}

static int get_client_MaxPoll(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "maxpoll", "10");
	return 0;
}

static int set_client_MaxPoll(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{NULL, NULL}}, 1))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "maxpoll", value);
		break;
	}

	return ret;
}

static int get_client_IBurst(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "iburst", value);
	return 0;
}

static int set_client_IBurst(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			ret = FAULT_9007;
		break;
	case VALUESET:
		string_to_bool(value, &b);
		dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "iburst", b ? "1" : "0");
		break;
	}

	return ret;
}

static int get_client_Interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *linker = NULL;

	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "interface", &linker);

	return bbf_get_reference_param("Device.IP.Interface.", "Name", linker, value);

}

static int set_client_Interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *allowed_objects[] = {"Device.IP.Interface.", NULL};
	struct dm_reference reference = {0};

	bbf_get_reference_args(value, &reference);

	switch (action)	{
		case VALUECHECK:
			if (dm_entry_validate_allowed_objects(ctx, reference.path, allowed_objects))
				return FAULT_9007;

			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "interface", reference.value);
			break;
	}
	return 0;	
}

static int get_client_PacketsSent(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	char *sec_name = section_name(((struct dmmap_dup *)data)->config_section);
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "client", String}, {"section", sec_name, String}}, 2, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsSent");
	}

	return 0;
}

static int get_client_PacketsSentFailed(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	char *sec_name = section_name(((struct dmmap_dup *)data)->config_section);
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "client", String}, {"section", sec_name, String}}, 2, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsSentFailed");
	}

	return 0;
}

static int get_client_PacketsReceived(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	char *sec_name = section_name(((struct dmmap_dup *)data)->config_section);
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "client", String}, {"section", sec_name, String}}, 2, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsReceived");
	}

	return 0;
}

static int get_client_PacketsDropped(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	char *sec_name = section_name(((struct dmmap_dup *)data)->config_section);
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "client", String}, {"section", sec_name, String}}, 2, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsDropped");
	}

	return 0;
}

static int get_time_ServerNumber(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	unsigned int cnt = get_number_of_entries(ctx, data, instance, browseTimeServer);
	dmasprintf(value, "%u", cnt);
	return 0;
}

static int get_server_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_option_value_fallback_def("time", "server", "enable", "1");
	return 0;
}

static int set_server_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			ret = FAULT_9007;
		break;
	case VALUESET:
		string_to_bool(value, &b);
		dmuci_set_value("time", "server", "enable", b ? "1" : "0");
		break;
	}

	return ret;
}

static int get_server_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "status", UBUS_ARGS{{"object", "server", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "Status");
	} else {
		*value = "Error";
	}

	return 0;
}

static int get_server_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, (struct uci_section *)data, "server_alias", instance, value);
}

static int set_server_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, (struct uci_section *)data, "server_alias", instance, value);
}

static int get_server_Mode(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_option_value_string("time", "server", "mode", value);
	return 0;
}

static int set_server_Mode(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;
	char *allowed_params[] = {"Unicast", "Broadcast", "Multicast", "Manycast", NULL};

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_string(ctx, value, -1, -1, allowed_params, NULL))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value("time", "server", "mode", value);
		break;
	}

	return ret;
}

static int get_server_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "123";
	return 0;
}

static int set_server_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1","65535"}}, 1))
			ret = FAULT_9007;

		if (DM_STRCMP(value, "123") != 0) {
			bbfdm_set_fault_message(ctx, "The current implementation supports only one port which is '123'. So the only allowed value is '123'.");
			ret = FAULT_9007;
		}

		break;
	case VALUESET:
		break;
	}

	return ret;
}

static int get_server_Interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *linker = NULL;

	dmuci_get_option_value_string("time", "server", "interface", &linker);

	return bbf_get_reference_param("Device.IP.Interface.", "Name", linker, value);
}

static int set_server_Interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *allowed_objects[] = {"Device.IP.Interface.", NULL};
	struct dm_reference reference = {0};

	bbf_get_reference_args(value, &reference);

	switch (action)	{
		case VALUECHECK:
			if (dm_entry_validate_external_linker_allowed_objects(ctx, reference.path, allowed_objects))
				return FAULT_9007;

			break;
		case VALUESET:
			dmuci_set_value("time", "server", "interface", reference.value);
			break;
	}
	return 0;	
}

static int get_server_TTL(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_option_value_fallback_def("time", "server", "ttl", "255");
	return 0;
}

static int set_server_TTL(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	int ret = 0;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1", "255"}}, 1))
			ret = FAULT_9007;
		break;
	case VALUESET:
		dmuci_set_value("time", "server", "ttl", value);
		break;
	}

	return ret;
}

static int get_server_PacketsSent(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "server", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsSent");
	}

	return 0;
}

static int get_server_PacketsSentFailed(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "server", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsSentFailed");
	}

	return 0;
}

static int get_server_PacketsReceived(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "server", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsReceived");
	}

	return 0;
}

static int get_server_PacketsDropped(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL;
	dmubus_call("timemngr", "stats", UBUS_ARGS{{"object", "server", String}}, 1, &res);
	if (res != NULL) {
		*value = dmjson_get_value(res, 1, "PacketsDropped");
	}

	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & LEAF DEFINITION
***********************************************************************************************************************************/
/*** Device. ***/
DMOBJ tDeviceObjs[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"Time", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tTimeObjs, tTimeParams, NULL, BBFDM_BOTH, NULL},
{0}
};

/* *** Device.Time. *** */
DMOBJ tTimeObjs[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"Client", &DMWRITE, addTimeClient, delTimeClient, NULL, browseTimeClient, NULL, NULL, tTimeClientStats, tTimeClientParams, NULL, BBFDM_BOTH, LIST_KEY{"Alias", NULL}},
{"Server", &DMWRITE, addTimeServer, delTimeServer, NULL, browseTimeServer, NULL, NULL, tTimeServerStats, tTimeServerParams, NULL, BBFDM_BOTH, LIST_KEY{"Alias", NULL}},
{0}
};

DMLEAF tTimeParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Enable", &DMWRITE, DMT_BOOL, get_time_Enable, set_time_Enable, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_time_Status, NULL, BBFDM_BOTH},
{"CurrentLocalTime", &DMREAD, DMT_TIME, get_time_CurrentLocalTime, NULL, BBFDM_BOTH},
{"LocalTimeZone", &DMWRITE, DMT_STRING, get_time_LocalTimeZone, set_time_LocalTimeZone, BBFDM_BOTH},
{"ClientNumberOfEntries", &DMREAD, DMT_UNINT, get_time_ClientNumber, NULL, BBFDM_BOTH},
{"ServerNumberOfEntries", &DMREAD, DMT_UNINT, get_time_ServerNumber, NULL, BBFDM_BOTH},
{0}
};

DMOBJ tTimeClientStats[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"Stats", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, NULL, tTimeClientStatsParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMOBJ tTimeServerStats[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"Stats", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, NULL, tTimeServerStatsParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tTimeClientParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"Enable", &DMWRITE, DMT_BOOL, get_client_Enable, set_client_Enable, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_client_Status, NULL, BBFDM_BOTH},
{"Alias", &DMWRITE, DMT_STRING, get_client_Alias, set_client_Alias, BBFDM_BOTH, DM_FLAG_UNIQUE},
{"Mode", &DMWRITE, DMT_STRING, get_client_Mode, set_client_Mode, BBFDM_BOTH},
{"Port", &DMWRITE, DMT_UNINT, get_client_Port, set_client_Port, BBFDM_BOTH},
{"Version", &DMWRITE, DMT_UNINT, get_client_Version, set_client_Version, BBFDM_BOTH},
{"Servers", &DMWRITE, DMT_STRING, get_client_Servers, set_client_Servers, BBFDM_BOTH},
{"Peer", &DMWRITE, DMT_BOOL, get_client_Peer, set_client_Peer, BBFDM_BOTH},
{"MinPoll", &DMWRITE, DMT_UNINT, get_client_MinPoll, set_client_MinPoll, BBFDM_BOTH},
{"MaxPoll", &DMWRITE, DMT_UNINT, get_client_MaxPoll, set_client_MaxPoll, BBFDM_BOTH},
{"IBurst", &DMWRITE, DMT_BOOL, get_client_IBurst, set_client_IBurst, BBFDM_BOTH},
{"Interface", &DMWRITE, DMT_STRING, get_client_Interface, set_client_Interface, BBFDM_BOTH, DM_FLAG_REFERENCE},
{0}
};

DMLEAF tTimeClientStatsParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"PacketsSent", &DMREAD, DMT_UNINT, get_client_PacketsSent, NULL, BBFDM_BOTH},
{"PacketsSentFailed", &DMREAD, DMT_UNINT, get_client_PacketsSentFailed, NULL, BBFDM_BOTH},
{"PacketsReceived", &DMREAD, DMT_UNINT, get_client_PacketsReceived, NULL, BBFDM_BOTH},
{"PacketsDropped", &DMREAD, DMT_UNINT, get_client_PacketsDropped, NULL, BBFDM_BOTH},
{0}
};

DMLEAF tTimeServerParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"Enable", &DMWRITE, DMT_BOOL, get_server_Enable, set_server_Enable, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_server_Status, NULL, BBFDM_BOTH},
{"Alias", &DMWRITE, DMT_STRING, get_server_Alias, set_server_Alias, BBFDM_BOTH, DM_FLAG_UNIQUE},
{"Mode", &DMWRITE, DMT_STRING, get_server_Mode, set_server_Mode, BBFDM_BOTH},
{"Port", &DMWRITE, DMT_UNINT, get_server_Port, set_server_Port, BBFDM_BOTH},
{"Interface", &DMWRITE, DMT_STRING, get_server_Interface, set_server_Interface, BBFDM_BOTH, DM_FLAG_REFERENCE},
{"TTL", &DMWRITE, DMT_UNINT, get_server_TTL, set_server_TTL, BBFDM_BOTH},
{0}
};

DMLEAF tTimeServerStatsParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"PacketsSent", &DMREAD, DMT_UNINT, get_server_PacketsSent, NULL, BBFDM_BOTH},
{"PacketsSentFailed", &DMREAD, DMT_UNINT, get_server_PacketsSentFailed, NULL, BBFDM_BOTH},
{"PacketsReceived", &DMREAD, DMT_UNINT, get_server_PacketsReceived, NULL, BBFDM_BOTH},
{"PacketsDropped", &DMREAD, DMT_UNINT, get_server_PacketsDropped, NULL, BBFDM_BOTH},
{0}
};
